//-----------------------------------------------------------------------------
// name: VisualSine.cpp
// desc: hello sine wave, real-time
//
// author: Ge Wang (ge@ccrma.stanford.edu)
//   date: fall 2014
//   uses: RtAudio by Gary Scavone
//-----------------------------------------------------------------------------
#include "RtAudio/RtAudio.h"
#include "chuck.h"
#include <math.h>
#include <stdlib.h>
#include <iostream>
using namespace std;

//#include "chuck_fft.h"

#ifdef __MACOSX_CORE__
#include <GLUT/glut.h>
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#endif




//-----------------------------------------------------------------------------
// function prototypes
//-----------------------------------------------------------------------------
void initGfx();
void idleFunc();
void displayFunc();
void reshapeFunc( GLsizei width, GLsizei height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );

// our datetype
#define SAMPLE float
// corresponding format for RtAudio
#define MY_FORMAT RTAUDIO_FLOAT32
// sample rate
#define MY_SRATE 44100
// number of channels
#define MY_CHANNELS_IN 1
#define MY_CHANNELS_OUT 2
#define GREATER(a,b) ( a > b ? a : b )
// for convenience
#define MY_PIE 3.14159265358979
//#define SCREEN_W = 2560.0
//#define SCREEN_H = 1600.0

// width and height
long g_width = 1024;
long g_height = 720;
// global buffer
SAMPLE * g_buffer = NULL;
long g_bufferSize;

// global variables
bool g_draw_dB = false;
ChucK * the_chuck = NULL;

int counter = 0;
//int singer = 0;
int numSingers = 36;
int* singers = new int [numSingers];

int starter = 0;
int currVolume = 0;



//-----------------------------------------------------------------------------
// name: callme()
// desc: audio callback
//-----------------------------------------------------------------------------
int callme( void * outputBuffer, void * inputBuffer, unsigned int numFrames,
           double streamTime, RtAudioStreamStatus status, void * data )
{
    // cast!
    SAMPLE * input = (SAMPLE *)inputBuffer;
    SAMPLE * output = (SAMPLE *)outputBuffer;
    
    the_chuck->run( input, output, numFrames );
    
    
    
    // fill
  
    for( int i = 0; i < numFrames; i++ )
    {
        g_buffer[i] = 0;
        for( int j = 0; j < MY_CHANNELS_OUT; j++ )
        {
            // accumulate each channel
            
            g_buffer[i] += output[i*MY_CHANNELS_OUT+j];
        }
        g_buffer[i] /= MY_CHANNELS_OUT;
     
        // copy the input to visualize only the left-most channel
        //g_buffer[i] = input[i*MY_CHANNELS_IN];
        
        // also copy in the output from chuck to our visualizer
        // (TODO: to fill in)
        
        // mute output -- TODO will need to disable this once ChucK produces output, in order for you to hear it!
        //for( int j = 0; j < MY_CHANNELS_OUT; j++ ) { output[i*MY_CHANNELS_OUT + j] = 0; }
      
    }
     
    
    
    
    return 0;
}




//-----------------------------------------------------------------------------
// name: main()
// desc: entry point
//-----------------------------------------------------------------------------
int main( int argc, char ** argv )
{
    // instantiate RtAudio object
    RtAudio audio;
    // variables
    unsigned int bufferBytes = 0;
    // frame size
    unsigned int bufferFrames = 1024;
    
    // check for audio devices
    if( audio.getDeviceCount() < 1 )
    {
        // nopes
        cout << "no audio devices found!" << endl;
        exit( 1 );
    }
    
    // initialize GLUT
    glutInit( &argc, argv );
    // init gfx
    initGfx();
    
    // let RtAudio print messages to stderr.
    audio.showWarnings( true );
    
    // set input and output parameters
    RtAudio::StreamParameters iParams, oParams;
    iParams.deviceId = audio.getDefaultInputDevice();
    iParams.nChannels = MY_CHANNELS_IN;
    iParams.firstChannel = 0;
    oParams.deviceId = audio.getDefaultOutputDevice();
    oParams.nChannels = MY_CHANNELS_OUT;
    oParams.firstChannel = 0;
    
    // create stream options
    RtAudio::StreamOptions options;
    
    // go for it
    try {
        // open a stream
        audio.openStream( &oParams, &iParams, MY_FORMAT, MY_SRATE, &bufferFrames, &callme, (void *)&bufferBytes, &options );
    }
    catch( RtError& e )
    {
        // error!
        cout << e.getMessage() << endl;
        exit( 1 );
    }
    
    // compute
    bufferBytes = bufferFrames * GREATER(MY_CHANNELS_OUT, MY_CHANNELS_IN) * sizeof(SAMPLE);
    // allocate global buffer
    g_bufferSize = bufferFrames;
    g_buffer = new SAMPLE[g_bufferSize];
    memset( g_buffer, 0, sizeof(SAMPLE)*g_bufferSize );
    
    // set up chuck
    the_chuck = new ChucK();
    the_chuck->setParam( CHUCK_PARAM_SAMPLE_RATE, MY_SRATE );
    the_chuck->setParam( CHUCK_PARAM_INPUT_CHANNELS, MY_CHANNELS_IN );
    the_chuck->setParam( CHUCK_PARAM_OUTPUT_CHANNELS, MY_CHANNELS_OUT );
    the_chuck->init();
    if( !the_chuck->compileFile( argv[1], "", 1 ) ) { exit( 1 ); }
    
    
    // go for it
    try {
        // start stream
        audio.startStream();
        
        // let GLUT handle the current thread from here
        glutMainLoop();
        
        // stop the stream.
        audio.stopStream();
    }
    catch( RtError& e )
    {
        // print error message
        cout << e.getMessage() << endl;
        goto cleanup;
    }
    
cleanup:
    // close if open
    if( audio.isStreamOpen() )
        audio.closeStream();
    
    // done
    return 0;
}




//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void initGfx()
{
    // double buffer, use rgb color, enable depth buffer
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    // initialize the window size
    glutInitWindowSize( g_width, g_height );
    // set the window postion
    glutInitWindowPosition( 100, 100 );
    // create the window
    glutCreateWindow( "VisualSine" );
    
    // set the idle function - called when idle
    glutIdleFunc( idleFunc );
    // set the display function - called when redrawing
    glutDisplayFunc( displayFunc );
    // set the reshape function - called when client area changes
    glutReshapeFunc( reshapeFunc );
    // set the keyboard function - called on keyboard events
    glutKeyboardFunc( keyboardFunc );
    // set the mouse function - called on mouse stuff
    glutMouseFunc( mouseFunc );
    
    // set clear color
    //glClearColor( cos(counter), cos(2*counter), cos(3*counter), 1 );
    glClearColor( 0,0,0, 1 );
    
    // enable color material
    glEnable( GL_COLOR_MATERIAL );
    // enable depth test
    glEnable( GL_DEPTH_TEST );
    
    glEnable( GL_TEXTURE_2D );
    
    for (int i=0; i<numSingers; i++) { singers[i] = -1; }
}




//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( GLsizei w, GLsizei h )
{
    // save the new window size
    g_width = w; g_height = h;
    // map the view port to the client area
    glViewport( 0, 0, w, h );
    // set the matrix mode to project
    glMatrixMode( GL_PROJECTION );
    // load the identity matrix
    glLoadIdentity( );
    // create the viewing frustum
    gluPerspective( 45.0, (GLfloat) w / (GLfloat) h, 1.0, 300.0 );
    // set the matrix mode to modelview
    glMatrixMode( GL_MODELVIEW );
    // load the identity matrix
    glLoadIdentity( );
    // position the view point
    gluLookAt( 0.0f, 0.0f, 10.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f );
    sleep(1);
}




//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
    switch( key )
    {
        case 'Q':
        case 'q':
            exit(1);
            break;
            
        case 'd':
            g_draw_dB = !g_draw_dB;
            break;
    }
    
    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: mouseFunc( )
// Desc: handles mouse stuff
//-----------------------------------------------------------------------------
void mouseFunc( int button, int state, int x, int y )
{
    if( button == GLUT_LEFT_BUTTON )
    {
        // when left mouse button is down
        if( state == GLUT_DOWN )
        {
        }
        else
        {
        }
    }
    else if ( button == GLUT_RIGHT_BUTTON )
    {
        // when right mouse button down
        if( state == GLUT_DOWN )
        {
        }
        else
        {
        }
    }
    else
    {
    }
    
    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
    // render the scene
    glutPostRedisplay( );
}



void makeCirc (float lineWidth, float m, float radius, float xDis, float yDis, int c1, int c2)
{
    glColor3f(c1,c2,0);
    glLineWidth(lineWidth);
    glBegin( GL_LINE_STRIP );
    for (int a=m; a<360; a++)
    {
        float piVal = (a*MY_PIE)/180.0;
        glVertex2f((radius*cos(piVal))+xDis, (radius*sin(piVal))+yDis);
        
    }
    glEnd();
}

void makeLine (float lineWidth, float radius, float xDis, float yDis, int c1, int c2)
{
    glColor3f(c1,c2,0);
    glLineWidth(lineWidth);
    glBegin( GL_LINE_STRIP );
    glVertex2f(xDis-radius, yDis);
    glVertex2f(xDis+radius, yDis);
    glEnd();
}

void sing (float radius, float xDis, float yDis)
{
    glColor3f(1,1,1);
    glLineWidth( 1.0 );
    
    glBegin( GL_LINE_STRIP );
    float deg = counter*MY_PIE/90;
    for( int i = 0; i < counter; i++ )
    {
        glVertex2f(xDis + radius*((deg/100.0)*cos(deg)*2 + g_buffer[i]*5), + yDis + radius*((deg/100.0)*sin(deg)*2 + g_buffer[i]*5));
        //xPlace += (xinc*m);
        deg = deg+0.2;
        //cout << to_string(g_buffer[i]) << " ";
    }
    //cout << endl;
    //cout << endl;
    glEnd();
    
}

float getAverage(SAMPLE* g_buffer)
{
    float total = 0;
    for (int i=0; i<g_bufferSize; i++)
    {
        total += abs(g_buffer[i]);
    }
    return (total/(2*g_bufferSize));
}

int find(int*singers, int num)
{
    for (int z=0; z<numSingers; z++)
    {
        if (singers[z] == num) { return -1;}
    }
    return 1;
}

void printVec()
{
    for (int z=0; z<numSingers; z++){
        cout << singers[z];
    }
    cout << endl;
}

//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{
    int width = 100;
    int height = 100;
    
    counter = (counter + 4)%g_bufferSize;
    //if (counter < 4) { singer = rand() % 36; }
    //glClearColor( cos(counter), cos(2*counter), cos(3*counter), 1 );
    static GLfloat zrot = 0.0f, c = 0.0f;
    
    // clear the color and depth buffers
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    
    float radius = 0.15;
    float xDis = 0;
    float yDis = 3.5;
    float space = 0.3;
    
    int volume = floor(35*getAverage(g_buffer))+1;
    if (volume > numSingers/9) { volume = numSingers/9; }
    if ((volume == 1) && (counter < 4)) {
        singers[0] = rand() % 36;
    }
    if (volume > currVolume)
    {
        for (int j=currVolume; j<volume; j++)
        {
            int curr = rand() % 36;
            while (find(singers, curr) == -1) { int curr = rand() % 36; }
            singers[j] = curr;
        }
        printVec;
        currVolume = volume;
    }
    else if (volume < currVolume)
    {
        for (int j=volume; j<numSingers; j++) { singers[j] = -1; }
        printVec;
        currVolume = volume;
    }

    
    for (int row = 0; row<6; row++)
    {
        xDis = -((5*radius)+(2.5*space));
        for (int col = 0; col<6; col++)
        {
            
            if (find(singers, (6*row)+col) == -1)
            {
                makeCirc(0.1, 0, radius, xDis, yDis, 0, 1);
                makeCirc(1, 180, radius*0.7, xDis, yDis, 0, 1);
                makeLine(1, radius*0.2, xDis-(radius/3.0), yDis+(radius/4.0)+0.05, 0, 1);
                makeLine(1, radius*0.2, xDis+(radius/3.0), yDis+(radius/4.0)+0.05, 0, 1);
                makeLine(1, (radius*0.7), xDis, yDis, 0, 1);
                sing (radius, xDis, yDis-(radius/2.8));
                //if (col<3) { ; }
                //else { sing (1, radius, xDis, yDis-(radius/2.8)); }
            } else {
                makeCirc(0.1, 0, radius, xDis, yDis, 1, 0);
                makeCirc(1, 180, radius*0.7, xDis, yDis, 1, 0);
                makeCirc(1, 0, radius*0.1, xDis-(radius/3.0), yDis+(radius/4.0), 1, 0);
                makeCirc(1, 0, radius*0.1, xDis+(radius/3.0), yDis+(radius/4.0), 1, 0);
            }
            xDis += (radius*2)+space;
        }
        yDis -= (2*radius)+(1.5*space);
        radius += 0.12;
    }
    
    
    // flush!
    glFlush( );
    // swap the double buffer
    glutSwapBuffers( );
}
